#$Header: /raid/cvsroot/DBIx/EasySearch/EasySearch.pm,v 1.20 2000/05/02 05:07:48 jesse Exp $

package DBIx::EasySearch;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;
require AutoLoader;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw(
	
);
$VERSION = '0.05';


# Preloaded methods go here.

# {{{ sub new 

#instantiate a new object.

sub new  {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  $self->_Init(@_);
  return ($self)
}
# }}}

# {{{ Methods dealing with Record objects

# {{{ sub Next 

sub Next  {
  my $self = shift;
  my @row;
  
  if (!$self->_isLimited) {
    return(0);
  }
  if ($self->{'must_redo_search'} != 0) {
    $self->_DoSearch();
  }
  
  if ($self->{'itemscount'} < $self->{'rows'}) {
    #increment the itemcounter.
    $self->{'itemscount'}++;
    #serve out that item
    return ($self->{'items'}[$self->{'itemscount'}]);
  }
  else {
    #we've gone through the whole list.
    #reset the count.
    $self->{'itemscount'} = 0;
    return(undef);
  }
}

# }}}

# {{{ sub First 

sub First  {
  my $self = shift;
  #Reset the itemcount
  $self->{'itemscount'} = 0;
  return ($self->Next);
}

# }}}

# {{{ sub NewItem 
sub NewItem  {
  my $self = shift;
  my $foo = (caller(1))[3];                                                        
  die "DBIx::EasySearch needs to be subclassed. you can't use it directly. ($foo)\n";
}
# }}}

# }}}

# {{{ sub UnLimit 
sub UnLimit {
  my $self=shift;
  $self->_isLimited(-1);
}

# }}} 

# {{{ sub Limit 

sub Limit  {
  my $self = shift;
  my %args = (
	      TABLE => $self->{'table'},
	      FIELD => undef,
	      VALUE => undef,
	      ALIAS => undef,
	      TYPE => undef, 
	      ENTRYAGGREGATOR => 'or',
	      INT_LINKFIELD => $self->{'primary_key'},
	      EXT_LINKFIELD => 'id',
	      OPERATOR => '=',
	      OFFSET => 0,
	      ORDERBY => undef,
	      ORDER => undef,
	      ROWS => undef,
	      @_ # get the real argumentlist
	     );
  
  my ($Alias);

  if ($args{'FIELD'}) {
    #If it's a like, we supply the %s around the search term
    if ($args{'OPERATOR'} eq "LIKE") {
      $args{'VALUE'} = "%".$args{'VALUE'} ."%";
    }
    $args{'VALUE'} = $self->_Handle->quote($args{'VALUE'});
  }
  
  $Alias = $self->_GenericRestriction(%args);
  
  # {{{ Set the order.
  # Do we want "ASC"ending order or "DESC"ending order
  if ($args{'ORDER'}) {
    $self->{'order'} = $args{'ORDER'};
  }

  # }}}
 
  # {{{ Set the limit clause.
  
  # LIMIT clauses are used for restricting ourselves to subsets of the search.
  if ( $args{'ROWS'}) {
      $self->{'limit_clause'} = "LIMIT ";
      if ($args{'OFFSET'} != 0) {
	  $self->{'limit_clause'} .= $args{'OFFSET'} . ", ";
      }
    $self->{'limit_clause'} .= $args{'ROWS'};
  }
  # }}}

  # {{{ If we're setting an OrderBy, set it here
  if ($args{'ORDERBY'}) {
    # 1. if an alias was passed in, use that. 
    if ($args{'ALIAS'}) {
      $self->{'orderby'} = $args{'ALIAS'}.".".$args{'ORDERBY'};
    }
    # 2. if an alias was generated, use that.
    elsif ($Alias) {
       $self->{'orderby'} = "$Alias.".$args{'ORDERBY'};
     }
    # 3. use the primary
    else {
      $self->{'orderby'} = "main.".$args{'ORDERBY'};
    }
  }
  # }}}

  # We're now limited. people can do searches.

  $self->_isLimited(1);
  
  if (defined ($Alias)) {
    return($Alias);
  }
  else {
    return(0);
  }
}
# }}}

# {{{ Routines dealing with Restrictions (where subclauses)

# {{{ sub ShowRestrictions 

#Show Restrictions
sub ShowRestrictions  {
   my $self = shift;
  $self->_CompileGenericRestrictions();
   $self->_CompileSubClauses();
  return($self->{'where_clause'});
  
}

# }}}

# {{{ sub ImportRestrictions 

#import a restrictions clause
sub ImportRestrictions  {
  my $self = shift;
  $self->{'where_clause'} = shift;
}
# }}}

# {{{ sub _GenericRestriction 
sub _GenericRestriction  {
  my $self = shift;
  my %args = (
	      TABLE => $self->{'table'},
	      FIELD => undef,
	      VALUE => undef,   #TODO: $Value should take an array of values and generate the proper where clause.
	      ALIAS => undef,	     
	      ENTRYAGGREGATOR => undef,
	      OPERATOR => '=',
	      INT_LINKFIELD => undef,
	      EXT_LINKFIELD => undef,
	      RESTRICTION_TYPE => 'generic', 
	      @_);
  my ($QualifiedField);
  
  #since we're changing the search criteria, we need to redo the search
  $self->{'must_redo_search'}=1;
  
  # {{{ if there's no alias set, we need to set it
  if (!$args{'ALIAS'}) {
    
    #if the table we're looking at is the same as the main table
    if ($args{'TABLE'} eq $self->{'table'}) {
      
      # main is the alias of the "primary table.
      # TODO this code assumes no self joins on that table. 
      # if someone can name a case where we'd want to do that, I'll change it.
      
      $args{'ALIAS'} = 'main';
    }
    
    # {{{ if we're joining, we need to work out the table alias
    else {
      $self->{'aliascount'}++;
      $args{'ALIAS'} = $args{'TABLE'}."_".$self->{'aliascount'};
      $self->{'auxillary_tables'} .= ",".$args{'TABLE'}." as $args{'ALIAS'}";
      
      # we need to build the table of links.
      $self->{'table_links'} .= "AND main.". $args{'INT_LINKFIELD'}."=".
	                          $args{'ALIAS'}.".".$args{'EXT_LINKFIELD'};
    }
    # }}}
  }
  # }}}
  
  # If we were just setting an alias, return
  if (!$args{'FIELD'}) {
    return ($args{'ALIAS'});
    
  }
  
  #Set this to the name of the field and the alias.
  $QualifiedField = $args{'ALIAS'}.".".$args{'FIELD'};
  print STDERR "DBIx::EasySearch->_GenericRestriction  QualifiedField is $QualifiedField\n" 
    if ($self->DEBUG);
  
  #If we're overwriting this sort of restriction, 
  if (($args{'ENTRYAGGREGATOR'} eq 'none') or 
      (!$self->{'restrictions'}{"$QualifiedField"})) {
    $self->{'restrictions'}{"$QualifiedField"} = 
      "($QualifiedField $args{'OPERATOR'} $args{'VALUE'})";  

  }
  else {
    $self->{'restrictions'}{"$QualifiedField"} .= 
      " $args{'ENTRYAGGREGATOR'} ($QualifiedField $args{'OPERATOR'} $args{'VALUE'})";

  }
  

  return ($args{'ALIAS'});
  
}
# }}}

# {{{ sub _AddRestriction
sub _AddSubClause {
    my $self = shift;
    my $clauseid = shift;
    my $subclause = shift;

    $self->{'subclauses'}{"$clauseid"} = $subclause;
    
}
# }}}

# {{{ sub _CompileSubClauses

sub _CompileSubClauses {
 my $self = shift;
  my ($subclause);
 $self->{'where_clause'} = undef;
  
 #Go through all restriction types. Build the where clause from the 
 #Various subclauses.
 foreach $subclause (keys %{ $self->{'subclauses'}}) {
     # Now, build up the where clause
     if (defined ($self->{'where_clause'})) {
	 $self->{'where_clause'} .= " AND ";
     }
     $self->{'where_clause'} .= $self->{'subclauses'}{"$subclause"};
 }
}

# }}}

# {{{ sub _CompileGenericRestrictions 

#Compile the restrictions to a WHERE Clause

sub _CompileGenericRestrictions  {
    my $self = shift;
    my ($restriction);
    $self->{'subclauses'}{'generic_restrictions'} = undef;
    
    #Go through all the restrictions of this type. Buld up the generic subclause
    foreach $restriction (keys %{ $self->{'restrictions'}}) {
	if (defined $self->{'subclauses'}{'generic_restrictions'}) {
	    $self->{'subclauses'}{'generic_restrictions'} .= " AND ";
	}
	$self->{'subclauses'}{'generic_restrictions'} .= 
	  "(" . $self->{'restrictions'}{"$restriction"} . ")";
    }
}

# }}}

# }}}

# {{{ routines dealing with table aliases

# {{{ sub NewAlias
sub NewAlias {
    #TODO Stub
    die "NewAlias stubbed";
    my $self = shift;
    my $table = shift;
    my ($alias);
    return ($alias);
}
# }}}

# {{{ sub _CompileAliases

sub _CompileAliases {
    my $self = shift;
    #TODO: Implement
    return ($self->{'compiled_aliases'});
}

# }}}


# things we'll want to add:
# get aliases
# add restirction clause
# 

# }}}

# {{{ sub _Order 

# This routine returns what the result set should be ordered by
sub _Order {

  my $self = shift;
  my $OrderClause;
  
  if ($self->{'orderby'}) {
    $OrderClause = " ORDER BY ". $self->{'orderby'};
  }
  else {
    $OrderClause = " ORDER BY main.".$self->{'primary_key'}." ";
  }
  return ($OrderClause);
}
# }}}

# {{{ sub _OrderBy 

# This routine returns whether the result set should be sorted Ascending or Descending

sub _OrderBy {
  my $self = shift;
  my $order_by;
  if ($self->{'order'} =~ /^des/) {
    $order_by = " DESC";
  }
  else {
    $order_by = " ASC";
  }
  return($order_by);
}

# }}}

# {{{ private utility methods

# {{{ sub _Init 

#Initialize the object

sub _Init  {
  my $self = shift;
  my %args = ( Handle => undef,
	       @_
	     );
  $self->{'DBIxHandle'} = $args{'Handle'}; 
  $self->{'must_redo_search'}=1;
  $self->{'itemscount'}=0;
  $self->{'tables'} = "";
  $self->{'auxillary_tables'} = "";
  $self->{'where_clause'} = "";
  $self->{'table_links'} = "";
  $self->{'limit_clause'} = "";
  $self->{'order'} = "";

  #we have no limit statements. DoSearch won't work.
  $self->_isLimited(0);
}

# }}}

# {{{ sub _Handle 
sub _Handle  {
  my $self = shift;
  return ($self->{'DBIxHandle'}->Handle);
}
# }}}

# {{{ sub _DoSearch 

sub _DoSearch  {
    my $self = shift;
    my ($QueryString, $Order);
    
    #compile restrictions into an intelligent where clause
    $self->_CompileGenericRestrictions();
    $self->_CompileSubClauses();
    # {{{ Build the query string
  $QueryString = "SELECT distinct main.".$self->{'primary_key'} ." FROM " .
    $self->{'table'}." AS main". $self->{'auxillary_tables'};

    $QueryString .= " WHERE ".
    $self->{'where_clause'}." ".$self->{'table_links'}." ".
	$self->{'limit_clause'}
    if ($self->_isLimited > 0);
  
  $QueryString .=  $self->_Order . $self->_OrderBy;
  
  # }}}
    
    print STDERR "DBIx::EasySearch->DoSearch Query:  $QueryString\n" 
      if ($self->DEBUG);
    
    # {{{ get $self->{'records'} out of the database
  $self->{'records'} = $self->_Handle->prepare($QueryString);
  
  if (!$self->{'records'}) {
    die "Error:" . $self->_Handle->errstr . "\n";
  }
  if (!$self->{'records'}->execute) {
    die "DBIx::EasySearch error:" . $self->{'records'}->errstr . "\n\tQuery String is $QueryString\n";
  }
    
  # }}}
    
    # populate $self->{'rows'} for $Count;
    $self->{'rows'} = $self->{'records'}->rows();
    
    my $counter = 0;
    
    # {{{ Iterate through all the rows returned and get child objects
  # TODO: this could be made much more efficient

  while (my @row = $self->{'records'}->fetchrow()) {
    $counter++;
      
    $self->{'items'}[$counter] = $self->NewItem();
        $self->{'items'}[$counter]->Load($row[0]);
    
    print "ID is ". $self->{'items'}[$counter]->Id()."\n"
      if ($self->DEBUG);

  }
  
  # }}}
    
    $self->{'must_redo_search'}=0;
    
    return($self->Count);
}

# }}}

# {{{ sub _isLimited 
sub _isLimited  {
    my $self = shift;
    if (@_) {
	$self->{'is_limited'} = shift;
    }
    else {
	return ($self->{'is_limited'});
    }
}
# }}}

# }}}

# {{{ Public utility methods

# {{{ sub DEBUG 
sub DEBUG {
    my $self = shift;
    if (@_) {
	$self->{'DEBUG'} = shift;
    }
    return ($self->{'DEBUG'});
    
}

# }}}

# {{{ sub Count 
sub Count  {
    my $self = shift;
    
    if ($self->{'must_redo_search'}) {
	return ($self->_DoSearch);
    }
    else {
	return($self->{'rows'});
    }
}
# }}}

# {{{ sub quote 

# A convenience method
sub quote {
  my $self = shift;
return ($self->_Handle->quote(@_));
}

# }}}

# }}}

# Autoload methods go after =cut, and are processed by the autosplit program.


1;
__END__

=head1 NAME

DBIx::EasySearch - Perl extension for easy SQL SELECT Statement generation

=head1 SYNOPSIS

  use DBIx::EasySearch;

  then read the code. (yes, i'm being lame)

=head1 DESCRIPTION

Jesse is lame and hasn't written docs yet

=head1 AUTHOR

Jesse Vincent, jesse@fsck.com

=head1 SEE ALSO

DBIx::Handle, DBIx::Record, perl(1).

=cut
